/*
==============================================================================
	lisp.c
		utility functions like LISP
		1990/11/12/Mon	Yutaka MYOKI(Nagao Lab., KUEE)
		1990/12/16/Mon	Last Modified
		special thanks to Itsuki NODA
==============================================================================
*/

#include "chadic.h"

#define COMMENTCHAR	';'
#define BPARENTHESIS	'('
#define BPARENTHESIS2	'<'
#define BPARENTHESIS3	'['
#define EPARENTHESIS	')'
#define EPARENTHESIS2	'>'
#define EPARENTHESIS3	']'
#define SCANATOM	"%[^(;) \n\t]"
#define NILSYMBOL	"NIL"
#define CELLALLOCSTEP	1024

#define _Car(cell)	(((cell_t *)(cell))->value.cons.car)
#define _Cdr(cell)	(((cell_t *)(cell))->value.cons.cdr)
#define new_cell()	(cons(NIL, NIL))
#define eq(x, y)	(x == y)

static char        Buffer[BUFSIZ];

/*
------------------------------------------------------------------------------
	local error processing
------------------------------------------------------------------------------
*/

static void error_in_lisp()
{
    cha_exit_file(DicError, "parse error");
}

static void error_in_program()
{
    cha_exit(ProgramError, "\"ifnextchar\" returns an unexpected code.");
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<ifnextchar>: if next char is <c> return 1, otherwise return 0
------------------------------------------------------------------------------
*/

static int ifnextchar(fp, i)
    FILE *fp;
    int i;
{
    int c;

    do {
	c = fgetc(fp);
	if (c == '\n') LineNo++;
    } while (c == ' ' || c == '\t' || c == '\n');

    if (c == EOF) return EOF;

    if (i == c) 
      return TRUE;
    else {
	ungetc(c, fp);
	return FALSE;
    }
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<s_feof>:
	after skipping "whitespace" , return TRUE if fp points to EOF
------------------------------------------------------------------------------
*/

static int s_feof_comment(fp)
    FILE *fp;
{
    int n;
    char Buffer[BUFSIZ];

    if ((n = ifnextchar(fp, (int)COMMENTCHAR)) == TRUE) {
	if (fgets(Buffer, BUFSIZ, fp) == NULL)
	  cha_exit_file(SystemError, "read error in comment");
	LineNo++;
	/* comment(fp); maybe bug. fixed by kurohashi. */
	return s_feof_comment(fp);
    }

    return n;
}

/* bug fixed by kurohashi on 92/4/25 */
int s_feof(fp)
    FILE *fp;
{
    int code, c;

    if (s_feof_comment(fp) == EOF) {
	code = TRUE;
    } else {
	if ((c = fgetc(fp)) == EOF) {
	    code = TRUE;
	} else {
	    if (c == '\n') {
		LineNo++;
		code = s_feof(fp);
	    } else if (c == ' ' || c == '\t') {
		code = s_feof(fp);
	    } else {
		ungetc(c, fp);
		code = FALSE;
	    }
	}
    }
    return code;
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<make_cell>: make a new cell
------------------------------------------------------------------------------
*/

static cell_t *make_cell()
{
    static int idx = CELLALLOCSTEP;
    static cell_t *ptr;

    if (idx == CELLALLOCSTEP) {
	ptr = (cell_t *)cha_malloc(sizeof(cell_t) * idx);
	idx = 0;
    }

    return (cell_t *)ptr + idx++;
}

#if 0
void free_cell(cell)
    cell_t *cell;
{
    if (atomp(cell)) {
	free(cell->value.atom);
    } else if (consp(cell)) {
	free_cell(car(cell));
	free_cell(cdr(cell));
    } else {
	return;
    }
    
    free(cell);
}
#endif

/*
------------------------------------------------------------------------------
	FUNCTION
	<tmp_atom>: use <TmpCell>
------------------------------------------------------------------------------
*/

cell_t *tmp_atom(atom)
    char *atom;
{
    static cell_t _TmpCell;
    static cell_t *TmpCell = &_TmpCell;

    s_tag(TmpCell) = ATOM;
    s_atom_val(TmpCell) = atom;

    return TmpCell;
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<cons>: make <cons> from <car> & <cdr>
------------------------------------------------------------------------------
*/

cell_t *cons(car, cdr)
    void *car, *cdr;
{
    cell_t *cell;

    cell = make_cell();
    s_tag(cell) = CONS;
    _Car(cell) = car;
    _Cdr(cell) = cdr;

    return cell;
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<car>: take <car> from <cons>
------------------------------------------------------------------------------
*/

cell_t *car(cell)
    cell_t *cell;
{
    if (consp(cell))
      return _Car(cell);
    else if (nullp(cell))
      return NIL;
    else {
	cha_exit_file(-1, "");
	s_print(stderr, cell);
	fprintf(stderr, " is not list\n");
	exit(1);
    }
    return NIL;
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<cdr>: take <cdr> from <cons>
------------------------------------------------------------------------------
*/

cell_t *cdr(cell)
    cell_t *cell;
{
    if (consp(cell))
      return _Cdr(cell);
    else if (nullp(cell))
      return NIL;
    else {
	s_print(stderr, cell);
	fprintf(stderr, " is not list. in <cdr>\n");
	error_in_lisp();
    }
    return NIL;
}

char *s_atom(cell)
    cell_t *cell;
{
    if (!atomp(cell)) {
	cha_exit_file(-1, "`");
	s_print(stderr, cell);
	fprintf(stderr, "' is not atom\n");
	exit(1);
    }

    return s_atom_val(cell);
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<equal>:
------------------------------------------------------------------------------
*/

int equal(x, y)
    void *x, *y;
{
    if (eq(x, y)) return TRUE;
    if (nullp(x) || nullp(y)) return FALSE;
    if (s_tag(x) != s_tag(y)) return FALSE;
    if (s_tag(x) == ATOM) return !strcmp(s_atom_val(x), s_atom_val(y));
    if (s_tag(x) == CONS)
      return (equal(_Car(x), _Car(y)) && equal(_Cdr(x), _Cdr(y)));
    return FALSE;
}

int s_length(list)
    cell_t *list;
{
    int i;

    for (i = 0; consp(list); i++) {
	list = _Cdr(list);
    }

    return i;
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<comment>: skip comment-line(s)
------------------------------------------------------------------------------
*/

static int comment(fp)
    FILE *fp;
{
    int n;

    if ((n = ifnextchar(fp, (int)COMMENTCHAR)) == TRUE) {
	if (fgets(Buffer, BUFSIZ, fp) == NULL)
	  cha_exit(1, "read error in comment");
	LineNo++;
	comment(fp);
    } else if (n == EOF) {
	/* error_in_lisp(); */
    }

    return n;
}

static int dividing_code_p(code)
    int code;
{
    switch (code) {
      case '\n': case '\t': case ';': 
      case '(': case ')': case ' ':      
	return 1;
      default:
	return 0;
    }
}

static int myscanf(fp, cp)
    FILE *fp;
    char *cp;
{
    int code;

    code = fgetc(fp);
    if ( dividing_code_p(code) )
      return 0;
    else if ( code == EOF )
      return EOF;
    else if ( code == '"' ) {
	*cp++ = code;
	while ( 1 ) {
	    code = fgetc(fp);
	    if ( code == EOF )
	      error_in_lisp();
	    else if ( code == '"' ) {
		*cp++ = code;
		*cp++ = '\0';
		return 1;
	    } 
	    else if ( code == '\\' ) {
		*cp++ = code;
		if ( (code = fgetc(fp)) == EOF ) 
		  error_in_lisp();
		*cp++ = code;
	    }	       
	    else {
		*cp++ = code;
	    }
	}
    }
    else {
	*cp++ = code;
	while ( 1 ) {
	    code = fgetc(fp);
	    if ( dividing_code_p(code) || code == EOF ) {
		*cp++ = '\0';
		ungetc(code, fp);
		return 1;
	    }
	    else {
		*cp++ = code;
	    }
	}
    }
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<s_read>: read S-expression
------------------------------------------------------------------------------
*/

static cell_t *s_read_atom(fp)
    FILE *fp;
{
    cell_t *cell;
    char *c;
    int n;

    comment(fp);

    /* changed by kurohashi. */
    if (((n = myscanf(fp, Buffer)) == 0) || (n == EOF)) {
	error_in_lisp();
    }

    if (!strcmp(Buffer, NILSYMBOL)) {
      cell = NIL;
    } else {
	cell = new_cell();
	s_tag(cell) = ATOM;
	c = cha_strdup(Buffer);
	s_atom_val(cell) = c;
    }

    return cell;
}

static cell_t *s_read_cdr();
static cell_t *s_read_main();

static cell_t *s_read_car(fp)
    FILE *fp;
{
    cell_t *cell;
    int n;

    comment(fp);

    if ((n = ifnextchar(fp, (int)EPARENTHESIS)) == TRUE) {
	cell = (cell_t *)NIL;
	return cell;
    } else if (n == FALSE) {
	cell = new_cell();
	_Car(cell) = s_read_main(fp);
	_Cdr(cell) = s_read_cdr(fp);
	return cell;
    }

    if (n == EOF) error_in_lisp();
    else          error_in_program();

    return NIL;
}

static cell_t *s_read_cdr(fp)
    FILE *fp;
{
    cell_t *cell;
    int n;

    comment(fp);
    if ((n = ifnextchar(fp, (int)EPARENTHESIS)) == TRUE) {
	cell = (cell_t *)NIL;
	return cell;
    } else if (n == FALSE) {
	cell = s_read_car(fp);
	return cell;
    }

    if (n == EOF) error_in_lisp();
    else          error_in_program();

    return NIL;
}

static cell_t *s_read_main(fp)
    FILE *fp;
{
    int n;

    if ((n = ifnextchar(fp, (int)BPARENTHESIS)) == TRUE) {
	return s_read_car(fp);
    } else if (n == FALSE) {
	return s_read_atom(fp);
    }

    if (n == EOF) error_in_lisp();
    else	   error_in_program();

    return NIL;
}

cell_t *s_read(fp)
    FILE *fp;
{
    if (LineNo == 0)
      LineNo = 1;
    LineNoForError = LineNo;

    return s_read_main(fp);
}

/*
------------------------------------------------------------------------------
	FUNCTION
	<assoc>:
------------------------------------------------------------------------------
*/

cell_t *assoc(item, alist)
    cell_t *item, *alist;
{
    for (; (!equal(item, (car(car(alist)))) && (!nullp(alist)));
	 alist = cdr(alist));
    return car(alist);
}

/*
------------------------------------------------------------------------------
	PROCEDURE
	<s_print>: pretty print S-expression
------------------------------------------------------------------------------
*/

static cell_t *s_print_cdr(fp, cell)
    FILE *fp;
    cell_t *cell;
{
    if (!nullp(cell)) {
	if (consp(cell)) {
	    fputc(' ', fp);
	    s_print(fp, _Car(cell));
	    s_print_cdr(fp, _Cdr(cell));
	} else {
	    fputc(' ', fp);
	    s_print(fp, cell);
	}
    }

    return cell;
}

cell_t *s_print(fp, cell)
    FILE *fp;
    cell_t *cell;
{
    if (nullp(cell))
      fputs(NILSYMBOL, fp);
    else {
	switch (s_tag(cell)) {
	  case CONS:
	    fprintf(fp, "%c", BPARENTHESIS);
	    s_print(fp, _Car(cell));
	    s_print_cdr(fp, _Cdr(cell));
	    fprintf(fp, "%c", EPARENTHESIS);
	    break;
	  case ATOM:
	    fprintf(fp, "%s", s_atom_val(cell));
	    break;
	  default:
	    cha_exit_file(1, "Illegal cell(in s_print)");
	}
    }

    return cell;
}

