/*
 * print.c - print mrphs and paths
 *
 * last modified by A.Kitauchi <akira-k@is.aist-nara.ac.jp>, Sep. 1996
 *               by O.Imaichi <osamu-im@is.aist-nara.ac.jp>, Sep. 1996
 */

#include "chasen.h"

#ifdef putc
#undef putc
#endif /* putc */

#define putc(c, fp)	chasend_putc(c)
#define fputs(s, fp)	chasend_puts(s)
#define printf		chasend_printf
#define cha_exit	chasend_error
#define BASIC_FORM 	"ܷ"

static short path_buffer[PATH_NUM];
static short p_check_buffer[PATH_NUM];

/***********************************************************************
 * printf_mrph - print morpheme using format string
 *
 * format string:
 *	%m    Ф(ѷ)
 *	%M    Ф(ܷ)
 *	%y    ɤ
 *	%Y    ɤ(̤ʤ"̤")
 *	%i    ̣
 *	%Ic   ̣(NILʤʸc)
 *	%Pc   ƳؤʻʸcǶڤäʸ(¿ʻΤ)
 *	%Pnc  1n(n:19)ܤޤǤʻʸcǶڤäʸ
 *            (¿ʻΤ)
 *	%h    ʻֹ
 *	%H    ʻʸ(¿ʻξ1)
 *	%Hn   n(n:19)ܤʻʸ(ʤкǤ⿼ؤɽ)
 *            (¿ʻΤ)
 *	%b    ʻʬֹ(¿ʻξ0)
 *	%BB   ʻʬʸ(ʤʻ)
 *	%BM   ʻʬʸ(ʤʻޤ"̤")
 *	%Bc   ʻʬʸ(ʤʸc)
 *	%t    ѷֹ
 *	%Tc   ѷʸ(ʤʸc)
 *	%f    ѷֹ
 *	%Fc   ѷʸ(ʤʸc)
 *	%c    ǤΥ
 *	%%    % Τ
 *	.     ե
 *	-     ե
 *	1-9   ե
 *	\n    ʸ
 *	\t    
 *	\\    \ Τ
 *	\'    ' Τ
 *	\"    " Τ
 *
 * example:
 *	"%m %y %M %h %b %t %f\n"                - same as '-c' option
 *	"%m %Y %M %H %h %B* %b %T* %t %F* %f\n" - same as '-e' option
 ***********************************************************************/
#ifdef VGRAM
static void print_nhinsi(hinsi, c, n)
    int hinsi, c, n;
{
    short *path;
    int i;

    path = Hinsi[hinsi].path;
    for (i = 0; ; i++) {
	fputs(Hinsi[*path].name, stdout);
	if (!*++path || i == n)
	  break;
	putc(c, stdout);
    }
}
#endif


/***********************************************************************
 * int_to_str - convert an integer to ASCII
 *	by Masanao Izumo <masana-i@is.aist-nara.ac.jp>
 ***********************************************************************/
static char *int_to_str(value)
    int value;
{
    static char buff[32];
    char* p;
    int sign;

    p = buff + 31;
    if (value >= 0)
      sign = 0;
    else {
	if (-value == value) { /* value == INT_MIN */
	    sprintf(buff, "%d", value);
	    return buff;
	}
	value = -value;
	sign = 1;
    }

    do {
	*--p = value % 10 + '0';
	value /= 10;
    } while (value > 0);
    if(sign)
      *--p = '-';

    return p;
}    

/***********************************************************************
 * fputsn
 *	by Masanao Izumo <masana-i@is.aist-nara.ac.jp>
 ***********************************************************************/
static void fputsn(str, fp, n)
    char *str;
    FILE* fp;
    int n;
{
    char buff[256];
    int len;

    while (n > 0) {
	len = (n <= 255 ? n : 255);
	memcpy(buff, str, len);
	buff[len] = '\0';
	fputs(buff, fp);
	str += len;
	n -= len;
    }
}

/***********************************************************************
 * printf_field
 *	by Masanao Izumo <masana-i@is.aist-nara.ac.jp>
 ***********************************************************************/
static void printf_field(width_str, word)
    char *width_str, *word;
{
    char *field = width_str;
    int field_len, word_len;

    if (width_str == NULL) {
	fputs(word, stdout);
	return;
    }

    if (*field == '-')
      field++;

    word_len = -1;
    field_len = 0;
    while('0' <= *field && *field <= '9') {
	field_len = field_len * 10 + *field - '0';
	field++;
    }

    if (*field == '.') {
	int len = 0;
	word_len = strlen(word);
	field++;
	while ('0' <= *field && *field <= '9') {
	    len = len * 10 + *field - '0';
	    field++;
	}
	if (len < word_len)
	  word_len = len;
    }

    if (word_len == -1) {
	word_len = strlen(word);
	if (*width_str == '-') {
	    fputs(word, stdout);
	    field_len -= word_len;
	    while(field_len-- > 0)
	      putc(' ', stdout);
	} else {
	    field_len -= word_len;
	    while(field_len-- > 0)
	      putc(' ', stdout);
	    fputs(word, stdout);
	}
    } else {
	if (*width_str == '-') {
	    fputsn(word, stdout, word_len);
	    field_len -= word_len;
	    while(field_len-- > 0)
	      putc(' ', stdout);
	} else {
	    field_len -= word_len;
	    while(field_len-- > 0)
	      putc(' ', stdout);
	    fputsn(word, stdout, word_len);
	}
    }
}

static void printf_mrph(mrph, format)
    mrph2_t *mrph;
    char *format;
{
    int  letter, value, n, c;
    char *s, *word, word_str[CHA_INPUT_SIZE];
    char *width_str;

    word = NULL;      /* string */
    letter = 0;       /* character */
    value = INT_MAX;  /* integer value */

    for (s = format; *s; s++) {
	/* escape characters */
	if (*s == '\\') {
	    switch (*++s) {
	      case 't':
		putc('\t', stdout);
		break;
	      case 'n':
		putc('\n', stdout);
		break;
	      case '\\':
	      case '\'':
	      case '"':
		putc(*s, stdout);
		break;
	      default:
		goto error_end;
	    }
	    continue;
	}

	if (*s != '%') {
	    putc(*s, stdout);
	    continue;
	}

	s++;
	if (*s == '-' || *s == '.' || (*s >= '0' && *s <= '9')) {
	    width_str = s;
	    while (*s == '-' || *s == '.' || (*s >= '0' && *s <= '9'))
	      s++;
	} else {
	    width_str = NULL;
	}

	switch (*s) {
	  case 'm': /* Ф(ѷ) */
	    memcpy(word_str, mrph->midasi, mrph->length);
	    word_str[mrph->length] = '\0';
	    word = word_str;
	    break;
	  case 'M': /* Ф(ܷ) */
	    memcpy(word_str, mrph->midasi, mrph->base_length);
	    if (mrph->ktype > 0 && mrph->kform > 0) {
		int b;
		for (b = 1;
		     strcmp(Form[mrph->ktype][b].name, BASIC_FORM);
		     b++);
		strcpy(word_str + mrph->base_length,
		       Form[mrph->ktype][b].gobi);
	    } else {
		word_str[mrph->base_length] = '\0';
	    }
	    word = word_str;
	    break;
	  case 'Y': /* ɤ / "̤" */
	    if (mrph->is_undef) {
		word = "̤";
		break;
	    }
	    /* don't break */
	  case 'y': /* ɤ */
	    strcpy(word_str, mrph->yomi);
	    if (mrph->ktype > 0 && mrph->kform > 0)
	      strcat(word_str, Form[mrph->ktype][mrph->kform].ygobi);
	    word = word_str;
	    break;
	  case 'i': /* ̣ */
	    word = mrph->info;
	    break;
	  case 'I': /* ̣ */
	    if (*++s == '\0')
	      goto error_end;
	    if (strcmp(mrph->info, "NIL"))
	      word = mrph->info;
	    else
	      letter = *s;
	    break;
#ifdef VGRAM
	  case 'P':
	    if (s[1] >= '1' && s[1] <= '9')
	      n = *++s - '1';
	    else
	      n = 99; /* ٤Ƥγؤɽ */
	    if (*++s == '\0')
	      goto error_end;
	    print_nhinsi(mrph->hinsi, *s, n);
	    break;
#endif
	  case 'h': /* ʻ(ֹ) */
	    value = mrph->hinsi;
	    break;
	  case 'H': /* ʻ(ʸ) */
#ifdef VGRAM
	    if (s[1] < '1' || s[1] > '9')
	      n = 0;
	    else {
		n = *++s - '1';
		if (Hinsi[mrph->hinsi].depth - 1 < n)
		  n = Hinsi[mrph->hinsi].depth - 1;
	    }
	    word = Hinsi[Hinsi[mrph->hinsi].path[n]].name;
#else
	    word = Class[mrph->hinsi][0].id;
#endif
	    break;
	  case 'b': /* ʻʬ(ֹ) */
#ifdef VGRAM
	    value = 0;
#else
	    value = mrph->bunrui;
#endif
	    break;
	  case 'B': /* ʻʬ(ʸ) */
	    if (*++s == '\0')
	      goto error_end;
	    if (*s == 'M' && mrph->is_undef)
	      word = "̤";
#ifdef VGRAM
	    /*
	     * زʻʤֲγؤʻ̾ɽ
	     * %BB  %Bc ζ̤Ϥʤ
	     */
	    else
	      word = Hinsi[mrph->hinsi].name;
#else
	    else if (*s == 'M' || *s == 'B' || mrph->bunrui)
	      word = Class[mrph->hinsi][mrph->bunrui].id;
	    else
	      letter = *s;
#endif
	    break;
	  case 't': /* ѷ(ֹ) */
	    value = mrph->ktype;
	    break;
	  case 'T': /* ѷ(ʸ) */
	    if (*++s == '\0')
	      goto error_end;
	    if (mrph->ktype)
	      word = Type[mrph->ktype].name;
	    else
	      letter = *s;
	    break;
	  case 'f': /* ѷ(ֹ) */
	    value = mrph->kform;
	    break;
	  case 'F': /* ѷ(ʸ) */
	    if (*++s == '\0')
	      goto error_end;
	    if (mrph->kform)
	      word = Form[mrph->ktype][mrph->kform].name;
	    else
	      letter = *s;
	    break;
	  case 'c': /* ǤΥ */
#ifdef VGRAM
	    value = Hinsi[mrph->hinsi].cost
	      * mrph->weight * Mrph_cost_weight;
#else
	    value = Class[mrph->hinsi][mrph->bunrui].cost
	      * mrph->weight * Mrph_cost_weight;
#endif
	    break;
	  case '%': /* '%'  */
	    putc(*s, stdout);
	    break;
	  default:
	    goto error_end;
	}

	if (word != NULL) {
	    printf_field(width_str, word);
	    word = NULL;
	} else if (letter) {
	    word_str[0] = letter;
	    word_str[1] = '\0';
	    printf_field(width_str, word_str);
	    letter = 0;
	} else if (value != INT_MAX) {
	    printf_field(width_str, int_to_str(value));
	    value = INT_MAX;
	}
    }

    return;

  error_end:
    cha_exit("500 format error near the point %d in option -- F", s - format + 1);
}

/***********************************************************************
 * print_mrph - print morpheme using option or format string
 ***********************************************************************/
void print_mrph(mrph, opt_form, format)
    mrph2_t *mrph;
    int opt_form;
    char *format;
{
    /* opt_form  'F' ʤ format 򤽤Τޤ޻ */
    switch (opt_form) {
#ifdef VGRAM
      case 'd':
	format = ",'%m','%Y','%M',%H1,%H2,%T0,%F0,'%I0',%c";
	break;
      case 'f':
	/* !VGRAM Ʊ */
	format = "%-11m %-11y %-11M %-14BM %-14T  %F \n";
	break;
      case 'e':
	format = "%m %Y %M %P- %h %T* %t %F* %f\n";
	break;
      case 'c':
	format = "%m %y %M %h %t %f\n";
	break;
#else
      case 'd':
	format = ",'%m','%Y','%M',%H,%BB,%T0,%F0,'%I0',%c";
	break;
      case 'f':
	format = "%-11m %-11y %-11M %-14BM %-14T  %F \n";
	break;
      case 'e':
	format = "%m %Y %M %H %h %B* %b %T* %t %F* %f\n";
	break;
      case 'c':
	format = "%m %y %M %h %b %t %f\n";
	break;
#endif
    }

    printf_mrph(mrph, format);
    if(opt_form != 'd')
	chasen_terminate_line();
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <print_path_mrph>
------------------------------------------------------------------------------
*/
static void print_path_mrph(path_num, opt_form, format)
    int path_num, opt_form;
    char *format;
{
#ifdef KOCHA_DEBUG
    printf("%4d: ", Path[path_num].mrph_p);
#endif
    print_mrph(&Mrph[Path[path_num].mrph_p], opt_form, format);
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <print_best_path> ǾȤΥѥΤߤɽ
------------------------------------------------------------------------------
*/
void print_best_path(opt_form, format)
    int opt_form;
    char *format;
{
    int i, j, last;
    mrph2_t mrph, *mrph1, *mrph2;
    char yomi[CHA_INPUT_SIZE];

    j = 0;
    last = Path_num - 1;
#if 0
    printf("cost: %d\n", Path[last].cost);
#endif
    do {
	last = Path[last].path[0];
	path_buffer[j] = last;
	j++;
    } while (Path[last].path[0]);

    /* ϢʻñϢ뤷ɽ */
    mrph.hinsi = 0;
    mrph.yomi = yomi;
    mrph1 = &Mrph[Path[path_buffer[j - 1]].mrph_p];
    for (i = j - 1; i >= 0; i--) {
	mrph2 = i == 0 ? NULL : &Mrph[Path[path_buffer[i - 1]].mrph_p];
	if (i > 0 &&
	    !mrph1->is_undef && !mrph2->is_undef &&
#ifdef VGRAM
	    Hinsi[mrph1->hinsi].comp &&
	    Hinsi[mrph1->hinsi].comp == Hinsi[mrph2->hinsi].comp
#else
	    Class[mrph1->hinsi][mrph1->bunrui].comp &&
	    mrph1->hinsi == mrph2->hinsi &&
	    (Class[mrph1->hinsi][mrph1->bunrui].comp == 1 ||
	     mrph1->bunrui == mrph2->bunrui)
#endif
	    ) {
	    if (!mrph.hinsi) {
#ifdef VGRAM
		mrph.hinsi = Hinsi[mrph1->hinsi].comp;
#else
		mrph.hinsi = mrph1->hinsi;
		if (Class[mrph1->hinsi][mrph1->bunrui].comp == 1)
		  mrph.bunrui = 0;
		else
		  mrph.bunrui = mrph1->bunrui;
#endif
		mrph.midasi = mrph1->midasi;
		yomi[0] = '\0';
		mrph.length = mrph.weight = 0;
	    }
	    strcat(yomi, mrph1->yomi);
	    if (mrph1->ktype > 0 && mrph1->kform > 0)
	      strcat(yomi, Form[mrph1->ktype][mrph1->kform].ygobi);
	    mrph.length += mrph1->length;
	    mrph.weight += mrph1->weight;
	} else {
	    if (!mrph.hinsi)
	      print_mrph(mrph1, opt_form, format);
	    else {
		strcat(yomi, mrph1->yomi);
		mrph.base_length = mrph.length + mrph1->base_length;
		mrph.length += mrph1->length;
		mrph.weight += mrph1->weight;
		mrph.info = mrph1->info;
		mrph.ktype = mrph1->ktype;
		mrph.kform = mrph1->kform;
		mrph.is_undef = mrph1->is_undef;
		print_mrph(&mrph, opt_form, format);
		mrph.hinsi = 0;
	    }
	}
	mrph1 = mrph2;
    }
#if 0
    for (i = j - 1; i >= 0; i--) {
	print_path_mrph(path_buffer[i], opt_form, format);
    }
#endif
}
/*
------------------------------------------------------------------------------
  PROCEDURE: <print_all_mrph> Ϸ̤˴ޤޤƤη
------------------------------------------------------------------------------
*/
static void collect_all_mrph(path_num)
    int path_num;
{
    int i, j;

    for (i = 0; (j = Path[path_num].path[i]) != -1; i++) {
	if (!p_check_buffer[j]) {
	    p_check_buffer[j] = 1;
	    collect_all_mrph(j);
	}
    }
}

static int check_con_cost(path, con_tbl)
    path_t *path;
    int con_tbl;
{
    int con_cost;

    check_automaton(path->state,
#ifdef KOCHA
		    Path[path->path[0]].state,
		    Mrph[path->mrph_p].midasi,
		    Mrph[path->mrph_p].is_undef,
#endif
		    con_tbl, Undef_con_cost, &con_cost);

    return con_cost;
}

void print_all_mrph(opt_form, format)
    int opt_form;
    char *format;
{
    int i, j;
    int isfirst = 1; /* ʸƬɤΥե饰 for -d option */

    memset(p_check_buffer, 0, sizeof(int) * Path_num);
    collect_all_mrph(Path_num - 1);

    switch (opt_form) {
      case 'd':
	for ( i=1; i<Path_num; i++ ) {
	    if (p_check_buffer[i]) {
		path_t *path = &Path[i];
		if (isfirst == 1) {
		    isfirst = 0;
		    fputs("[", stdout);
		} else {
		    fputs(",", stdout);
		}

		/* IDϰ֡λ֡ѥ */
		printf("morph(%d,%d,%d,%d",
			i, path->start, path->end, path->cost);
		print_path_mrph(i, opt_form, format);
		/* ܷǤIDȤȤϢܥ */
		if (path->path[0] == -1)
		  printf(",[],[]");
		else {
		    int con_tbl = Mrph[path->mrph_p].con_tbl;
		    printf(",[%d", Con_cost_weight *
			    check_con_cost(&Path[path->path[0]], con_tbl));
		    for (j = 1; path->path[j] != -1; j++)
		      printf(",%d", Con_cost_weight *
			      check_con_cost(&Path[path->path[j]], con_tbl));
		    printf("]");
		    printf(",[%d", path->path[0]);
		    for (j = 1; path->path[j] != -1; j++)
		      printf(",%d", path->path[j]);
		    printf("])");
		}
	    }
	}
	printf("].");
	chasen_terminate_line();
	break;
      default:
	for (i = 1; i < Path_num; i++)
	  if (p_check_buffer[i])
	    print_path_mrph(i, opt_form, format);
	break;
    }
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <print_all_path> Ϸ̤˴ޤޤƤPATH
------------------------------------------------------------------------------
*/
static void print_all_path_sub(path_num, paths, opt_form, format)
    int path_num, paths, opt_form;
    char *format;
{
    int i, j;

    for (i = 0; Path[path_num].path[i] != -1; i++) {
	if (Path[path_num].path[0] == 0) {
	    for (j = paths - 1; j >= 0; j--)
	      print_path_mrph(path_buffer[j], opt_form, format);
	    printf("EOP\n");
	} else {
	    path_buffer[paths] = Path[path_num].path[i];
	    print_all_path_sub(Path[path_num].path[i], paths + 1,
			       opt_form, format);
	}
    }
}

void print_all_path(opt_form, format)
    int opt_form;
    char *format;
{
    print_all_path_sub(Path_num-1, 0, opt_form, format);
}
